/*
Licence
-------
The code in this repository is available under the [MIT License](https://secure.wikimedia.org/wikipedia/en/wiki/Mit_license).
Copyright (c) 2008-2012 Memo Akten, [www.memo.tv](http://www.memo.tv)
The Mega Super Awesome Visuals Company
*/

//ported from c++ (ofxMSAFluid) to SuperCollider by redFrik

MSAFluidDrawer {
	var fluidSolver; //MSAFluidSolver
	var alphaEnabled; //Boolean
	var drawMode; //Symbol
	var enabled= true;
	//var useAdditiveBlending= false; //not implemented in the sc port
	var brightness= 1.0;
	var doInvert= false;
	//var vectorSkipCount= 0; //not implemented in the sc port
	*new {
		^super.new.initMSAFluidDrawer;
	}
	initMSAFluidDrawer {
		this.enableAlpha(false);
		this.setDrawMode(\color);
	}
	setup {|solver|
		fluidSolver= solver;
	}
	getFluidSolver {
		^fluidSolver;
	}
	enableAlpha {|b|
		alphaEnabled= b;
	}
	reset {
		fluidSolver.reset;
	}
	update {
		fluidSolver.update;
	}
	setDrawMode {|newDrawMode|
		drawMode= newDrawMode;
	}
	incDrawMode {
		switch(drawMode,
			\color, {drawMode= \motion},
			\motion, {drawMode= \speed},
			\speed, {drawMode= \vectors},
			\vectors, {drawMode= \count},
			\count, {drawMode= \color}
		);
	}
	decDrawMode {
		switch(drawMode,
			\color, {drawMode= \count},
			\motion, {drawMode= \color},
			\speed, {drawMode= \motion},
			\vectors, {drawMode= \speed},
			\count, {drawMode= \vectors}
		);
	}
	getDrawMode {
		^drawMode; //here different in the sc port
	}
	getDrawModeName {
		^drawMode;
	}
	draw {|x, y, width, height|
		if(enabled, {
			switch(drawMode,
				\color, {this.drawColor(x, y, width, height)},
				\motion, {this.drawMotion(x, y, width, height)},
				\speed, {this.drawSpeed(x, y, width, height)},
				\vectors, {this.drawVectors(x, y, width, height)}
			);
		});
	}
	drawColor {|x, y, width, height, withAlpha= false|
		var fw, fh, vel, color, i, j= 1, index;
		var r, g, b;
		var sw, sh;
		fw= fluidSolver.getWidth;
		fh= fluidSolver.getHeight;
		sw= width/(fw-2);
		sh= height/(fh-2);
		while({j<(fh-1)}, {
			i= 1;
			while({i<(fw-1)}, {
				index= fluidSolver.getIndexForCell(i, j);
				vel= fluidSolver.getVelocityAtIndex(index);
				color= fluidSolver.getColorAtIndex(index);
				r= min(color[0]*brightness, 1.0);
				g= min(color[1]*brightness, 1.0);
				b= min(color[2]*brightness, 1.0);
				if(doInvert, {
					r= 1-r;
					g= 1-g;
					b= 1-b;
				});
				if(alphaEnabled and:{withAlpha}, {
					Pen.fillColor= Color(r, g, b, max(b, max(r, g)));
				}, {
					Pen.fillColor= Color(r, g, b);
				});
				Pen.fillRect(Rect(i-1*sw, j-1*sh, sw, sh));
				i= i+1;
			});
			j= j+1;
		});
	}
	drawMotion {|x, y, width, height, withAlpha= false|
		var fw, fh, vel, color, i, j= 1, index;
		var r, g;
		var speed;
		var sw, sh;
		fw= fluidSolver.getWidth;
		fh= fluidSolver.getHeight;
		sw= width/(fw-2);
		sh= height/(fh-2);
		while({j<(fh-1)}, {
			i= 1;
			while({i<(fw-1)}, {
				index= fluidSolver.getIndexForCell(i, j);
				vel= fluidSolver.getVelocityAtIndex(index);
				color= fluidSolver.getColorAtIndex(index);
				r= vel[0].abs*fw;
				g= vel[1].abs*fh;
				speed= min(r+g*brightness, 1.0);
				if(alphaEnabled and:{withAlpha}, {
					Pen.fillColor= Color(min(r*brightness, 1.0), min(g*brightness, 1.0), 0, speed);
				}, {
					Pen.fillColor= Color(min(r*brightness, 1.0), min(g*brightness, 1.0), 0);
				});
				Pen.fillRect(Rect(i-1*sw, j-1*sh, sw, sh));
				i= i+1;
			});
			j= j+1;
		});
	}
	drawSpeed {|x, y, width, height, withAlpha= false|
		var fw, fh, vel, color, i, j= 1, index;
		var r, g;
		var speed;
		var sw, sh;
		fw= fluidSolver.getWidth;
		fh= fluidSolver.getHeight;
		sw= width/(fw-2);
		sh= height/(fh-2);
		while({j<(fh-1)}, {
			i= 1;
			while({i<(fw-1)}, {
				index= fluidSolver.getIndexForCell(i, j);
				vel= fluidSolver.getVelocityAtIndex(index);
				color= fluidSolver.getColorAtIndex(index);
				r= vel[0].abs*fw;
				g= vel[1].abs*fh;
				speed= min(r+g*brightness, 1.0);
				if(alphaEnabled and:{withAlpha}, {
					Pen.fillColor= Color(speed, speed, speed, speed);
				}, {
					Pen.fillColor= Color(speed, speed, speed);
				});
				Pen.fillRect(Rect(i-1*sw, j-1*sh, sw, sh));
				i= i+1;
			});
			j= j+1;
		});
	}
	drawVectors {} //not implemented in sc port TODO
}

//version optimised for Image
MSAFluidDrawer2 : MSAFluidDrawer {
	var <>img;
	drawColor {|x, y, width, height, withAlpha= false|
		var fw, fh, vel, color, i, j= 1, index;
		var r, g, b, c= Color.new;
		fw= fluidSolver.getWidth;
		fh= fluidSolver.getHeight;
		while({j<(fh-1)}, {
			i= 1;
			while({i<(fw-1)}, {
				index= fluidSolver.getIndexForCell(i, j);
				vel= fluidSolver.getVelocityAtIndex(index);
				color= fluidSolver.getColorAtIndex(index);
				r= min(color[0]*brightness, 1.0);
				g= min(color[1]*brightness, 1.0);
				b= min(color[2]*brightness, 1.0);
				if(doInvert, {
					r= 1-r;
					g= 1-g;
					b= 1-b;
				});
				c.red= r;
				c.green= g;
				c.blue= b;
				if(alphaEnabled and:{withAlpha}, {
					c.alpha= max(b, max(r, g));
				});
				img.setColor(c, i, j);
				i= i+1;
			});
			j= j+1;
		});
	}
	drawMotion {|x, y, width, height, withAlpha= false|
		var fw, fh, vel, i, j= 1, index;
		var r, g, c= Color.new;
		var speed;
		fw= fluidSolver.getWidth;
		fh= fluidSolver.getHeight;
		while({j<(fh-1)}, {
			i= 1;
			while({i<(fw-1)}, {
				index= fluidSolver.getIndexForCell(i, j);
				vel= fluidSolver.getVelocityAtIndex(index);
				r= vel[0].abs*fw;
				g= vel[1].abs*fh;
				speed= min(r+g*brightness, 1.0);
				c.red= min(r*brightness, 1.0);
				c.green= min(g*brightness, 1.0);
				if(alphaEnabled and:{withAlpha}, {
					c.alpha= speed;
				});
				img.setColor(c, i, j);
				i= i+1;
			});
			j= j+1;
		});
	}
	drawSpeed {|x, y, width, height, withAlpha= false|
		var fw, fh, vel, i, j= 1, index;
		var r, g, c= Color.new;
		var speed;
		fw= fluidSolver.getWidth;
		fh= fluidSolver.getHeight;
		while({j<(fh-1)}, {
			i= 1;
			while({i<(fw-1)}, {
				index= fluidSolver.getIndexForCell(i, j);
				vel= fluidSolver.getVelocityAtIndex(index);
				r= vel[0].abs*fw;
				g= vel[1].abs*fh;
				speed= min(r+g*brightness, 1.0);
				c.red= speed;
				c.green= speed;
				c.blue= speed;
				if(alphaEnabled and:{withAlpha}, {
					c.alpha= speed;
				});
				img.setColor(c, i, j);
				i= i+1;
			});
			j= j+1;
		});
	}
}
