/*
Licence
-------
The code in this repository is available under the [MIT License](https://secure.wikimedia.org/wikipedia/en/wiki/Mit_license).
Copyright (c) 2008-2012 Memo Akten, [www.memo.tv](http://www.memo.tv)
The Mega Super Awesome Visuals Company
*/

//ported from c++ (ofxMSAFluid) to SuperCollider by redFrik

MSAFluidSolver {
	var density, densityOld;//FloatArray - used if not RGB
	var color, colorOld;//Array of FloatArray[0, 0, 0] - used for RGB
	var uv, uvOld;//Array of FloatArray[0, 0]
	var curl;//FloatArray
	var >doRGB;//for monochrome, update only density
	var doVorticityConfinement;
	var >solverIterations;
	var colorDiffusion;
	var viscocity;
	var >fadeSpeed;
	var deltaT;
	var wrap_x;
	var wrap_y;
	var <width, <height, <invWidth, <invHeight;
	var nx, nx1, nx2, ny, ny1, ny2;
	var nx1nx2, nx2ny, nx2ny1, nxnx2ny1;
	var numCells, invNumCells, numCells1;
	var avgDensity;//this will hold the average color of the last frame (how full it is)
	var uniformity;//this will hold the uniformity of the last frame (how uniform the color is)
	var avgSpeed;
	*new {|nx= 100, ny= 100, deltaT= 0.04, viscocity= 0.0001, fadeSpeed= 0.03, iterations= 10, colorDiffusion= 0|
		^super.new.setup(nx, ny, deltaT, viscocity, fadeSpeed, iterations, colorDiffusion);
	}
	setSize {|argNx, argNy|
		nx= argNx;
		ny= argNy;
		nx1= nx+1;
		ny1= ny+1;
		nx2= nx+2;
		ny2= ny+2;
		nx1nx2= nx1+nx2;
		nx2ny= nx2*ny;
		nx2ny1= nx2*ny1;
		nxnx2ny1= nx+nx2ny1;
		numCells= nx2*ny2;
		invNumCells= 1/numCells;
		numCells1= numCells-1;
		width= this.getWidth;
		height= this.getHeight;
		invWidth= 1/width;
		invHeight= 1/height;
		this.reset;
	}
	setup {|argNx, argNy, argDeltaT, argViscocity, argFadeSpeed, argIterations, argColorDiffusion|
		this.setDeltaT(argDeltaT);
		this.setFadeSpeed(argFadeSpeed);
		this.setSolverIterations(argIterations);
		this.enableVorticityConfinement(false);
		this.setWrap(false, false);
		viscocity= argViscocity;
		colorDiffusion= argColorDiffusion;
		this.setSize(argNx, argNy);
	}

	setDeltaT {|delta|
		deltaT= delta;
	}
	setFadeSpeed {|speed|
		fadeSpeed= speed;
	}
	setSolverIterations {|iterations|
		solverIterations= iterations;
	}
	enableRGB {|bool|
		doRGB= bool;
	}
	enableVorticityConfinement {|bool|
		doVorticityConfinement= bool;
	}
	getVorticityConfinement {
		^doVorticityConfinement;
	}
	setWrap {|boolx, booly|
		wrap_x= boolx;
		wrap_y= booly;
	}

	//clear all forces in fluid and reset
	reset {
		density= FloatArray.fill(numCells, {0});
		densityOld= FloatArray.fill(numCells, {0});
		color= Array.fill(numCells, {FloatArray[0, 0, 0]});
		colorOld= Array.fill(numCells, {FloatArray[0, 0, 0]});
		uv= Array.fill(numCells, {FloatArray[0, 0]});
		uvOld= Array.fill(numCells, {FloatArray[0, 0]});
		curl= FloatArray.fill(numCells, {0});
	}

	getNumCells {
		^numCells;
	}
	getWidth {
		^nx2;
	}
	getHeight {
		^ny2;
	}
	getInvWidth {
		^invWidth;
	}
	getInvHeight {
		^invHeight;
	}
	getSize {
		^FloatArray[this.getWidth, this.getHeight];
	}
	getInvSize {
		^FloatArray[invWidth, invHeight];
	}
	setVisc {|newVisc|
		viscocity= newVisc;
	}
	getVisc {
		^viscocity;
	}
	setColorDiffusion {|diff|
		colorDiffusion= diff;
	}
	getColorDiffusion {
		^colorDiffusion;
	}
	getAvgDensity {
		^avgDensity;
	}
	getUniformity {
		^uniformity;
	}
	getAvgSpeed {
		^avgSpeed;
	}

	calcCurl {|i, j|
		var nx2j= nx2*j;
		var du_dy= uv[i+(nx2*(j+1))][0]-uv[i+(nx2*(j-1))][0];
		var dv_dx= uv[i+1+nx2j][1]-uv[i-1+nx2j][1];
		^(du_dy-dv_dx)*0.5;
	}

	vorticityConfinement {|vc_xy|
		var i, j;
		var dw_dx, dw_dy, length, v;
		var nx2j, inx2j;
		j= ny;
		while({j>0}, {
			i= nx;
			while({i>0}, {
				curl[i+(nx2*j)]= this.calcCurl(i, j).abs;
				i= i-1;
			});
			j= j-1;
		});
		j= ny-1;
		while({j>1}, {
			i= nx-1;
			while({i>1}, {
				nx2j= nx2*j;
				//Find derivative of the magnitude (_N = del |w|)
				dw_dx= curl[i+1+nx2j]-curl[i-1+nx2j];
				dw_dy= curl[i+(nx2*(j+1))]-curl[i+(nx2*(j-1))];
				//Calculate vector length. (|_N|)
				//Add small factor to prevent divide by zeros
				length= ((dw_dx*dw_dx)+(dw_dy*dw_dy)).sqrt+0.000001;
				//N = (_N/|_N|)
				length= 2/length;
				dw_dx= dw_dx*length;
				dw_dy= dw_dy*length;
				v= this.calcCurl(i, j);
				//N x w
				inx2j= i+nx2j;
				vc_xy[inx2j][0]= dw_dy*v.neg;
				vc_xy[inx2j][1]= dw_dx*v;
				i= i-1;
			});
			j= j-1;
		});
		^vc_xy;
	}

	//solve one step of the fluid solver
	update {
		uv= this.addSource(uv, uvOld);
		if(doVorticityConfinement, {
			uvOld= this.vorticityConfinement(uvOld);
			uv= this.addSource(uv, uvOld);
		});
		this.swapUV;
		this.diffuseUV(viscocity);
		#uv, uvOld= this.project(uv, uvOld);
		this.swapUV;
		uv= this.advect2d(uv, uvOld);
		#uv, uvOld= this.project(uv, uvOld);
		if(doRGB, {
			color= this.addSource(color, colorOld);
			this.swapColor;
			if(colorDiffusion!=0 and:{deltaT!=0}, {
				this.diffuseRGB(0, colorDiffusion);
				this.swapColor;
			});
			this.advectRGB(0, uv);
			this.fadeRGB;
		}, {
			density= this.addSource(density, densityOld);
			this.swapDensity;
			if(colorDiffusion!=0 and:{deltaT!=0}, {
				density= this.diffuse(0, density, densityOld, colorDiffusion);
				this.swapDensity;
			});
			density= this.advect(0, density, densityOld, uv);
			this.fadeDensity;
		});
	}

	fadeDensity {
		var i;
		var holdAmount= 1-fadeSpeed;
		var totalDeviations, currentDeviation;
		var tmp;
		var uvi0, uvi1;
		avgDensity= 0;
		totalDeviations= 0;
		avgSpeed= 0;
		i= numCells1;
		while({i>=0}, {
			//clear old values
			uvOld[i]= FloatArray[0, 0];
			densityOld[i]= 0;
			//calc avg speed
			uvi0= uv[i][0];
			uvi1= uv[i][1];
			avgSpeed= avgSpeed+((uvi0*uvi0)+(uvi1*uvi1));
			//calc avg density
			tmp= min(1, density[i]);
			avgDensity= avgDensity+tmp;//add it up
			//calc deviation (for uniformity)
			currentDeviation= tmp-avgDensity;
			totalDeviations= totalDeviations+(currentDeviation*currentDeviation);
			//fade out old
			density[i]= tmp*holdAmount;
			i= i-1;
		});
		avgDensity= avgDensity*invNumCells;
		avgSpeed= avgSpeed*invNumCells;
		uniformity= 1/(1+(totalDeviations*invNumCells));//0: very wide distribution, 1: very uniform
	}
	fadeRGB {
		var i;
		var holdAmount= 1-fadeSpeed;
		var totalDeviations, currentDeviation;
		var density, tmp= FloatArray[0, 0, 0];
		var uvi0, uvi1;
		avgDensity= 0;
		totalDeviations= 0;
		avgSpeed= 0;
		i= numCells1;
		while({i>=0}, {
			//clear old values
			uvOld[i]= FloatArray[0, 0];
			colorOld[i]= FloatArray[0, 0, 0];
			//calc avg speed
			uvi0= uv[i][0];
			uvi1= uv[i][1];
			avgSpeed= avgSpeed+((uvi0*uvi0)+(uvi1*uvi1));
			//calc avg density
			tmp[0]= min(1, color[i][0]);
			tmp[1]= min(1, color[i][1]);
			tmp[2]= min(1, color[i][2]);
			density= max(tmp[0], max(tmp[1], tmp[2]));
			avgDensity= avgDensity+density;//add it up
			//calc deviation (for uniformity)
			currentDeviation= density-avgDensity;
			totalDeviations= totalDeviations+(currentDeviation*currentDeviation);
			//fade out old
			color[i]= tmp*holdAmount;
			i= i-1;
		});
		avgDensity= avgDensity*invNumCells;
		avgSpeed= avgSpeed*invNumCells;
		uniformity= 1/(1+(totalDeviations*invNumCells));//0: very wide distribution, 1: very uniform
	}

	advect {|bound, d, d0, duv|
		var i, j;
		var i0, j0, i1, j1;
		var x, y, s0, t0, s1, t1;
		var index;
		var dt0x= deltaT*nx;
		var dt0y= deltaT*ny;
		var nx2j0, nx2j1;
		j= ny;
		while({j>0}, {
			i= nx;
			while({i>0}, {
				index= i+(nx2*j);
				x= i-(dt0x*duv[index][0]);
				y= j-(dt0y*duv[index][1]);
				x= x.clip(0.5, nx+0.5);
				i0= x.asInteger;
				i1= i0+1;
				y= y.clip(0.5, ny+0.5);
				j0= y.asInteger;
				j1= j0+1;
				s1= x-i0;
				s0= 1-s1;
				t1= y-j0;
				t0= 1-t1;
				nx2j0= nx2*j0;
				nx2j1= nx2*j1;
				d[index]=
				(s0*((t0*d0[i0+nx2j0])+(t1*d0[i0+nx2j1])))+
				(s1*((t0*d0[i1+nx2j0])+(t1*d0[i1+nx2j1])));
				i= i-1;
			});
			j= j-1;
		});
		d= this.setBoundary(bound, d);
		^d;
	}
	advect2d {|uv, duv|
		var i, j;
		var i0, j0, i1, j1;
		var x, y, s0, t0, s1, t1;
		var index;
		var dt0x= deltaT*nx;
		var dt0y= deltaT*ny;
		var nx2j1;
		var i0nx2j0, i1nx2j0, t1duvi0nx2j1, t1duvi1nx2j1;
		j= ny;
		while({j>0}, {
			i= nx;
			while({i>0}, {
				index= i+(nx2*j);
				x= i-(dt0x*duv[index][0]);
				y= j-(dt0y*duv[index][1]);
				x= x.clip(0.5, nx+0.5);
				i0= x.asInteger;
				i1= i0+1;
				y= y.clip(0.5, ny+0.5);
				j0= y.asInteger;
				j1= j0+1;
				s1= x-i0;
				s0= 1-s1;
				t1= y-j0;
				t0= 1-t1;
				nx2j1= nx2*j1;
				i0nx2j0= i0+(nx2*j0);
				i1nx2j0= i1+(nx2*j0);
				t1duvi0nx2j1= t1*duv[i0+nx2j1];
				t1duvi1nx2j1= t1*duv[i1+nx2j1];
				uv[index][0]=
				(s0*((t0*duv[i0nx2j0][0])+(t1duvi0nx2j1[0])))+
				(s1*((t0*duv[i1nx2j0][0])+(t1duvi1nx2j1[0])));
				uv[index][1]=
				(s0*((t0*duv[i0nx2j0][1])+(t1duvi0nx2j1[1])))+
				(s1*((t0*duv[i1nx2j0][1])+(t1duvi1nx2j1[1])));
				i= i-1;
			});
			j= j-1;
		});
		uv= this.setBoundary2d(1, uv);
		uv= this.setBoundary2d(2, uv);
		^uv;
	}
	advectRGB {|bound, duv|
		var i, j;
		var i0, j0;
		var x, y, s0, t0, s1, t1;
		var index;
		var dt0x= deltaT*nx;
		var dt0y= deltaT*ny;
		j= ny;
		while({j>0}, {
			i= nx;
			while({i>0}, {
				index= i+(nx2*j);
				x= i-(dt0x*duv[index][0]);
				y= j-(dt0y*duv[index][1]);
				x= x.clip(0.5, nx+0.5);
				i0= x.asInteger;
				y= y.clip(0.5, ny+0.5);
				j0= y.asInteger;
				s1= x-i0;
				s0= 1-s1;
				t1= y-j0;
				t0= 1-t1;
				i0= i0+(nx2*j0);
				j0= i0+nx2;
				color[index]=
				(((colorOld[i0]*t0)+(colorOld[j0]*t1))*s0)+
				(((colorOld[i0+1]*t0)+(colorOld[j0+1]*t1))*s1);
				i= i-1;
			});
			j= j-1;
		});
		this.setBoundaryRGB;
	}

	diffuse {|bound, c, c0, diff|
		var a= deltaT*diff*nx*ny;
		c= this.linearSolver(bound, c, c0, a, 1+(4*a));
		^c;
	}
	diffuseRGB {|bound, diff|
		var a= deltaT*diff*nx*ny;
		this.linearSolverRGB(a, 1+(4*a));
	}
	diffuseUV {|diff|
		var a= deltaT*diff*nx*ny;
		this.linearSolverUV(a, 1+(4*a));
	}

	project {|xy, pDiv|
		var i, j;
		var h, index;
		var fx= 0.5*nx;
		var fy= 0.5*ny;
		h= -0.5/nx;
		j= ny;
		while({j>0}, {
			index= nx+(nx2*j);
			i= nx;
			while({i>0}, {
				pDiv[index][0]= h*(xy[index+1][0]-xy[index-1][0]+xy[index+nx2][1]-xy[index-nx2][1]);
				pDiv[index][1]= 0;
				index= index-1;
				i= i-1;
			});
			j= j-1;
		});
		//this.setBoundary02d(pDiv.collect{|x| FloatArray[x[0], 0]}).do{|x, i| pDiv.put(i, FloatArray[x[0], pDiv[i][1]])};
		//this.setBoundary02d(pDiv.collect{|x| FloatArray[x[1], 0]}).do{|x, i| pDiv.put(i, FloatArray[pDiv[i][0], x[0]])};
		pDiv= this.setBoundary02d(pDiv);
		pDiv= this.linearSolverProject(pDiv);
		j= ny;
		while({j>0}, {
			index= nx+(nx2*j);
			i= nx;
			while({i>0}, {
				xy[index][0]= xy[index][0]-(fx*(pDiv[index+1][0]-pDiv[index-1][0]));
				xy[index][1]= xy[index][1]-(fy*(pDiv[index+nx2][0]-pDiv[index-nx2][0]));
				index= index-1;
				i= i-1;
			});
			j= j-1;
		});
		xy= this.setBoundary2d(1, xy);
		xy= this.setBoundary2d(2, xy);
		^[xy, pDiv];
	}
	linearSolver {|bound, x, x0, a, c|
		var i, j, k;
		var index;
		c= 1/c;
		k= solverIterations;
		while({k>0}, {
			j= ny;
			while({j>0}, {
				index= nx+(nx2*j);
				i= nx;
				while({i>0}, {
					x[index]= ((x[index-1]+x[index+1]+x[index-nx2]+x[index+nx2])*a+x0[index])*c;
					index= index-1;
					i= i-1;
				});
				j= j-1;
			});
			x= this.setBoundary(bound, x);
			k= k-1;
		});
		^x;
	}
	linearSolverProject {|pdiv|
		var i, j, k;
		var index;
		var prev;
		k= solverIterations;
		while({k>0}, {
			j= ny;
			while({j>0}, {
				index= nx+(nx2*j);
				prev= pdiv[index+1][0];
				i= nx;
				while({i>0}, {
					prev= (pdiv[index-1][0]+prev+pdiv[index-nx2][0]+pdiv[index+nx2][0]+pdiv[index][1])*0.25;
					pdiv[index][0]= prev;
					index= index-1;
					i= i-1;
				});
				j= j-1;
			});
			//this.setBoundary02d(pdiv.collect{|x| FloatArray[x[0], 0]}).do{|x, i| pdiv.put(i, FloatArray[x[0], pdiv[i][1]])};
			pdiv= this.setBoundary02d(pdiv);
			k= k-1;
		});
		^pdiv;
	}
	linearSolverRGB {|a, c|
		var i, j, k;
		var index3, index4, index;
		c= 1/c;
		k= solverIterations;
		while({k>0}, {
			j= ny;
			while({j>0}, {
				index= nx+(nx2*j);
				index3= index-nx2;
				index4= index+nx2;
				i= nx;
				while({i>0}, {
					color[index]= ((color[index-1]+color[index+1]+color[index3]+color[index4])*a+colorOld[index])*c;
					index= index-1;
					index3= index3-1;
					index4= index4-1;
					i= i-1;
				});
				j= j-1;
			});
			this.setBoundaryRGB;
			k= k-1;
		});
	}
	/*
	linearSolverUV {|a, c|
	var i, j, k;
	var index;
	var localUV= uv.copy, localOldUV= uvOld.copy;//???TODO restrict here?
	var prevU, prevV;
	c= 1/c;
	k= solverIterations;
	while({k>0}, {
	j= ny;
	while({j>0}, {
	index= nx+(nx2*j);
	prevU= localUV[index+1][0];
	prevV= localUV[index+1][1];
	i= nx;
	while({i>0}, {
	prevU= ((localUV[index-1][0]+prevU+localUV[index-nx2][0]+localUV[index+nx2][0])*a+localOldUV[index][0])*c;
	prevV= ((localUV[index-1][1]+prevV+localUV[index-nx2][1]+localUV[index+nx2][1])*a+localOldUV[index][1])*c;
	localUV[index][0]= prevU;
	localUV[index][1]= prevV;
	index= index-1;
	i= i-1;
	});
	j= j-1;
	});
	uv= this.setBoundary2d(1, uv);
	k= k-1;
	});
	}
	*/
	linearSolverUV {|a, c|
		var i, j, k;
		var index;
		var prevU, prevV;
		c= 1/c;
		k= solverIterations;
		while({k>0}, {
			j= ny;
			while({j>0}, {
				index= nx+(nx2*j);
				prevU= uv[index+1][0];
				prevV= uv[index+1][1];
				i= nx;
				while({i>0}, {
					prevU= ((uv[index-1][0]+prevU+uv[index-nx2][0]+uv[index+nx2][0])*a+uvOld[index][0])*c;
					prevV= ((uv[index-1][1]+prevV+uv[index-nx2][1]+uv[index+nx2][1])*a+uvOld[index][1])*c;
					uv[index]= FloatArray[prevU, prevV];
					index= index-1;
					i= i-1;
				});
				j= j-1;
			});
			uv= this.setBoundary2d(1, uv);
			k= k-1;
		});
	}
	setBoundary {|bound, x|
		var i;
		var dst1, dst2, src1, src2;
		dst1= nx2;
		dst2= nx1nx2;
		if(wrap_x, {
			src2= 1+nx2;
			src1= nx+nx2;
		}, {
			src1= 1+nx2;
			src2= nx+nx2;
		});
		if(bound==1 and:{wrap_x.not}, {
			i= ny;
			while({i>0}, {
				x[dst1]= x[src1].neg;
				dst1= dst1+nx2;
				src1= src1+nx2;
				x[dst2]= x[src2].neg;
				dst2= dst2+nx2;
				src2= src2+nx2;
				i= i-1;
			});
		}, {
			i= ny;
			while({i>0}, {
				x[dst1]= x[src1];
				dst1= dst1+nx2;
				src1= src1+nx2;
				x[dst2]= x[src2];
				dst2= dst2+nx2;
				src2= src2+nx2;
				i= i-1;
			});
		});
		dst1= 1;
		dst2= 1+nx2ny1;
		if(wrap_y, {
			src2= 1+nx2;
			src1= 1+nx2ny;
		}, {
			src1= 1+nx2;
			src2= 1+nx2ny;
		});
		if(bound==2 and:{wrap_y.not}, {
			i= nx;
			while({i>0}, {
				x[dst1]= x[src1].neg;
				dst1= dst1+1;
				src1= src1+1;
				x[dst2]= x[src2].neg;
				dst2= dst2+1;
				src2= src2+1;
				i= i-1;
			});
		}, {
			i= nx;
			while({i>0}, {
				x[dst1]= x[src1];
				dst1= dst1+1;
				src1= src1+1;
				x[dst2]= x[src2];
				dst2= dst2+1;
				src2= src2+1;
				i= i-1;
			});
		});
		x[0]= 0.5*(x[1]+x[nx2]);
		x[nx2ny1]= 0.5*(x[1+nx2ny1]+x[nx2ny]);
		x[nx1]= 0.5*(x[nx]+x[nx1nx2]);
		x[nx1+nx2ny1]= 0.5*(x[nxnx2ny1]+x[nx1+nx2ny]);
		^x;
	}
	setBoundary02d {|x|
		var i;
		var dst1, dst2, src1, src2;
		dst1= nx2;
		dst2= nx1nx2;
		if(wrap_x, {
			src2= 1+nx2;
			src1= nx+nx2;
		}, {
			src1= 1+nx2;
			src2= nx+nx2;
		});
		i= ny;
		while({i>0}, {
			x[dst1][0]= x[src1][0];
			dst1= dst1+nx2;
			src1= src1+nx2;
			x[dst2][0]= x[src2][0];
			dst2= dst2+nx2;
			src2= src2+nx2;
			i= i-1;
		});
		dst1= 1;
		dst2= 1+nx2ny1;
		if(wrap_y, {
			src2= 1+nx2;
			src1= 1+nx2ny;
		}, {
			src1= 1+nx2;
			src2= 1+nx2ny;
		});
		i= nx;
		while({i>0}, {
			x[dst1]= x[src1];
			dst1= dst1+1;
			src1= src1+1;
			x[dst2]= x[src2];
			dst2= dst2+1;
			src2= src2+1;
			i= i-1;
		});
		x[0][0]= 0.5*(x[1][0]+x[nx2][0]);
		x[nx2ny1][0]= 0.5*(x[1+nx2ny1][0]+x[nx2ny][0]);
		x[nx1][0]= 0.5*(x[nx][0]+x[nx1nx2][0]);
		x[nx1+nx2ny1][0]= 0.5*(x[nxnx2ny1][0]+x[nx1+nx2ny][0]);
		^x;
	}
	setBoundary2d {|bound, xy|
		var i;
		var dst1, dst2, src1, src2;
		var bound1= bound-1;
		dst1= nx2;
		dst2= nx1nx2;
		if(wrap_x, {
			src2= 1+nx2;
			src1= nx+nx2;
		}, {
			src1= 1+nx2;
			src2= nx+nx2;
		});
		if(bound==1 and:{wrap_x.not}, {
			i= ny;
			while({i>0}, {
				xy[dst1][0]= xy[src1][0].neg;
				dst1= dst1+nx2;
				src1= src1+nx2;
				xy[dst2][0]= xy[src2][0].neg;
				dst2= dst2+nx2;
				src2= src2+nx2;
				i= i-1;
			});
		}, {
			i= ny;
			while({i>0}, {
				xy[dst1][0]= xy[src1][0];
				dst1= dst1+nx2;
				src1= src1+nx2;
				xy[dst2][0]= xy[src2][0];
				dst2= dst2+nx2;
				src2= src2+nx2;
				i= i-1;
			});
		});
		dst1= 1;
		dst2= 1+nx2ny1;
		if(wrap_y, {
			src2= 1+nx2;
			src1= 1+nx2ny;
		}, {
			src1= 1+nx2;
			src2= 1+nx2ny;
		});
		if(bound==2 and:{wrap_y.not}, {
			i= nx;
			while({i>0}, {
				xy[dst1][1]= xy[src1][1].neg;
				dst1= dst1+1;
				src1= src1+1;
				xy[dst2][1]= xy[src2][1].neg;
				dst2= dst2+1;
				src2= src2+1;
				i= i-1;
			});
		}, {
			i= nx;
			while({i>0}, {
				xy[dst1][1]= xy[src1][1];
				dst1= dst1+1;
				src1= src1+1;
				xy[dst2][1]= xy[src2][1];
				dst2= dst2+1;
				src2= src2+1;
				i= i-1;
			});
		});
		xy[0][bound1]= 0.5*(xy[1][bound1]+xy[nx2][bound1]);
		xy[nx2ny1][bound1]= 0.5*(xy[1+nx2ny1][bound1]+xy[nx2ny][bound1]);
		xy[nx1][bound1]= 0.5*(xy[nx][bound1]+xy[nx1nx2][bound1]);
		xy[nx1+nx2ny1][bound1]= 0.5*(xy[nxnx2ny1][bound1]+xy[nx1+nx2ny][bound1]);
		^xy;
	}
	setBoundaryRGB {
		var i;
		var dst1, dst2, src1, src2;
		dst1= nx2;
		dst2= nx1nx2;
		if(wrap_x, {
			src2= 1+nx2;
			src1= nx+nx2;
		}, {
			src1= 1+nx2;
			src2= nx+nx2;
		});
		i= ny;
		while({i>0}, {
			color[dst1]= color[src1];
			dst1= dst1+nx2;
			src1= src1+nx2;
			color[dst2]= color[src2];
			dst2= dst2+nx2;
			src2= src2+nx2;
			i= i-1;
		});
		dst1= 1;
		dst2= 1+nx2ny1;
		if(wrap_y, {
			src2= 1+nx2;
			src1= 1+nx2ny;
		}, {
			src1= 1+nx2;
			src2= 1+nx2ny;
		});
		i= nx;
		while({i>0}, {
			color[dst1]= color[src1];
			dst1= dst1+1;
			src1= src1+1;
			color[dst2]= color[src2];
			dst2= dst2+1;
			src2= src2+1;
			i= i-1;
		});
	}

	randomizeColor {
		var i, j;
		var index;
		i= this.getWidth-1;
		while({i>0}, {
			j= this.getHeight-1;
			while({j>0}, {
				index= i+(nx2*j);
				color[index]= FloatArray[1.0.rand, 1.0.rand, 1.0.rand];
				density[index]= 1.0.rand;
				j= j-1;
			});
			i= i-1;
		});
	}

	swapUV {
		var tmp= uv;
		uv= uvOld;
		uvOld= tmp;
	}
	swapColor {
		var tmp= color;
		color= colorOld;
		colorOld= tmp;
	}
	swapDensity {
		var tmp= density;
		density= densityOld;
		densityOld= tmp;
	}

	//--get index
	getIndexForCell {|i, j|
		i= i.clip(1, nx);
		j= j.clip(1, ny);
		^i+(nx2*j);
	}
	getIndexForPos {|pos|
		^this.getIndexForCell(floor(pos[0]*width), floor(pos[1]*height));
	}

	//--get info
	getInfoAtIndex {|index|
		//here sc port different - returns an array with vel and color
		^[this.getVelocityAtIndex(index), this.getColorAtIndex(index)];
	}
	getInfoAtCell {|i, j|
		^this.getInfoAtIndex(this.getIndexForCell(i, j));
	}
	getInfoAtPos {|pos|
		^this.getInfoAtIndex(this.getIndexForPos(pos));
	}

	//--get velocity
	getVelocityAtIndex {|index|
		^uv[index];
	}
	getVelocityAtCell {|i, j|
		^this.getVelocityAtIndex(this.getIndexForCell(i, j));
	}
	getVelocityAtPos {|pos|
		^this.getVelocityAtIndex(this.getIndexForPos(pos));
	}

	//--get color
	getColorAtIndex {|index|
		if(doRGB, {
			^color[index].copy;
		}, {
			^FloatArray[density[index], density[index], density[index]];
		});
	}
	getColorAtCell {|i, j|
		^this.getColorAtIndex(this.getIndexForCell(i, j));
	}
	getColorAtPos {|pos|
		^this.getColorAtIndex(this.getIndexForPos(pos));
	}

	//--add force
	addForceAtIndex {|index, force|
		uv[index]= uv[index]+force;
	}
	addForceAtCell {|i, j, force|
		this.addForceAtIndex(this.getIndexForCell(i, j), force);
	}
	addForceAtPos {|pos, force|
		this.addForceAtIndex(this.getIndexForPos(pos), force);
	}

	//--add color
	addColorAtIndex {|index, argColor|
		if(doRGB, {
			colorOld[index]= colorOld[index]+argColor;
		}, {
			density[index]= density[index]+argColor[0];
		});
	}
	addColorAtCell {|i, j, argColor|
		this.addColorAtIndex(this.getIndexForCell(i, j), argColor);
	}
	addColorAtPos {|pos, argColor|
		this.addColorAtIndex(this.getIndexForPos(pos), argColor);
	}
	addSource {|x, x0|
		var i, j, n1= x[0].size-1;
		i= numCells1;
		while({i>=0}, {
			j= n1;
			while({j>=0}, {
				x[i].put(j, x[i][j]+(x0[i][j]*deltaT));
				j= j-1;
			});
			i= i-1;
		});
		^x;
	}
}
